tf_device = '/gpu:1'

import torch

import torch.nn.functional as F

def SOLO_device(x,auto_resize=False,stupid_mode=False):
	torch.cuda.empty_cache() #########################################################################################################################
	if (not stupid_mode) and (not hasattr(SOLO_device, 'device')):
		import tensorflow as tf
		print("first initialization of physical device")
		with tf.device(tf_device):
			SOLO_device.device = tf.keras.models.load_model('./digital_twin_ref.h5')
		# SOLO_device.device.summary()
	# 
	# if multiple samples, recurse
	# if len(x.shape) > 4: return torch.stack([SOLO_device(xx) for xx in x])
	# well let's disable that, it could become debugging hell...
	# 
	if auto_resize: x = F.interpolate(x, size=(128,128), mode='bilinear')
	if not x.shape[-3:] == torch.Size([1,128,128]): raise Exception(f"invalid shape {x.shape} : in SOLO_device(x) we must have x.shape=[1,128,128] or x.shape=[N,1,128,128]")
	print(f"SOLO processing {x.shape}")
	if stupid_mode: # for debugging
		y = x
	else:
		import tensorflow as tf
		with tf.device(tf_device):
			y = torch.permute(x, (0, 3, 2, 1)) # (N,1,128,128) → (N,128,128,1) ; TODO who cares in principle ?
			y = SOLO_device.device(y.cpu().detach().numpy()).numpy()
			y = torch.tensor(y).to(x.device)
			y = torch.permute(y, (0, 3, 2, 1)) # (N,128,128,1) → (N,1,128,128) ; TODO who cares in principle ?
	return y

def SOLO_scale(im,margin=[0,0,0,0],res=[128,128],mode="bilinear"):
	if not isinstance(margin, list): margin = [margin]*4
	from torchvision.transforms.functional import resize
	from torchvision.transforms import InterpolationMode as interpol
	size = (res[0]-margin[2]-margin[3],res[1]-margin[0]-margin[1])
	if   mode=="nearest" : im = resize(im, size=size, interpolation=interpol.NEAREST)
	elif mode=="bilinear": im = resize(im, size=size, interpolation=interpol.BILINEAR)
	elif mode=="bicubic" : im = resize(im, size=size, interpolation=interpol.BICUBIC)
	elif mode=="lanczos" : im = resize(im, size=size, interpolation=interpol.LANCZOS)
	elif mode=="hamming" : im = resize(im, size=size, interpolation=interpol.HAMMING)
	elif mode=="box"     : im = resize(im, size=size, interpolation=interpol.BOX)
	# return F.pad(im, margin, "constant", 0)
	# return F.pad(im, margin, "reflect", 0)
	return F.pad(im, margin, "replicate", 0)

def SOLO_conv(im,ker=torch.tensor([[[[1.0]]]]),stride=None):
	ker = ker.to(im.device)
	if stride == None: stride = ker.shape[-2:]
	# padding = ker.shape[-2:]
	return F.conv_transpose2d(im, ker, stride=stride, padding=0, output_padding=0, dilation=1)

import random
class HistIterator:
	def __init__(self,tensors,step,shuffle=False):
		self.tensors = tensors
		self.step = step
		self.indices = []
		for i,t in enumerate(tensors):
			for j in range(t.shape[0]):
				self.indices.append([i,j])
		if shuffle: random.shuffle(self.indices)
	def __iter__(self):
		return self
	def get_from_indices(self,indices):
		return self.tensors[indices[0]][indices[1]]
	def __next__(self):
		step = min(self.step,len(self.indices))
		if step==0: raise StopIteration
		out = [self.get_from_indices(self.indices.pop(0)) for s in range(step)]
		return torch.stack(out)

def debug_HistIterator():
	# Example usage
	my_list = [
		torch.arange(0, 5),
		torch.arange(5, 7),
		torch.arange(7, 8),
		torch.arange(8, 10),
		torch.arange(10, 30),
		torch.arange(30, 37),
	]
	for t1,t2 in zip(HistIterator(my_list,5,shuffle=False),HistIterator(my_list,5,shuffle=True)): print(torch.hstack((t1,t2)))
	my_list = [
		torch.randn(11, 1,128,128),
		torch.randn(12, 1,128,128),
		torch.randn(3, 1,128,128),
		torch.randn(6, 1,128,128),
		torch.randn(7, 1,128,128),
		torch.randn(2, 1,128,128),
		torch.randn(22, 1,128,128),
		torch.randn(8, 1,128,128),
	]
	# for t1 in HistIterator(my_list,5,shuffle=False): print(t1.shape)
	for t1,t2 in zip(HistIterator(my_list,5,shuffle=False),HistIterator(my_list,5,shuffle=True)): print(t1.shape,t2.shape)

